#include <avr/io.h>
#include <avr/interrupt.h>
#define F_CPU 16000000
#include <util/delay.h>
#include <string.h>
#include <stdio.h>
#include <avr/sleep.h>
#include <avr/wdt.h>

#define FCC(c1,c2,c3,c4)	(((DWORD)c4<<24)+((DWORD)c3<<16)+((WORD)c2<<8)+(BYTE)c1)	/* FourCC */


#include "diskio.h"
#include "pff.h"

#define CTRL_BUT_DDR	DDRC
#define CTRL_BUT_PORT	PORTC
#define CTRL_BUT_PIN	PINC

#define CTRL_BUT_PLAY	PC6
#define CTRL_BUT_REC	PC0

#define REC_LED_DDR		DDRD
#define REC_LED_PORT	PORTD
#define REC_LED			PD7


#define PLAY_BUT_PRESSED()	(!(CTRL_BUT_PIN & (1<<CTRL_BUT_PLAY)))
#define REC_BUT_PRESSED()	(!(CTRL_BUT_PIN & (1<<CTRL_BUT_REC)))

#define SET_REC_LED()		OCR2A=255 //(REC_LED_PORT |= (1<<REC_LED))
#define CLR_REC_LED()		OCR2A=0 //(REC_LED_PORT &= ~(1<<REC_LED))

#define WUMETER()  OCR2A=(OCR1A>128) ? (OCR1A-128) : (128-OCR1A)

#define WRITE_BUF_SIZE	256

/*---------------------------------------------------------*/
/* Work Area                                               */
/*---------------------------------------------------------*/

unsigned char Buff1[256];		/* Wave output FIFO */
unsigned char Buff2[256];		/* Wave output FIFO */
//unsigned char Header[256];
unsigned char *buf1;
unsigned char *buf2;

WORD rb;			/* Return value. Put this here to avoid bugs of avr-gcc */



FATFS file_system;			/* File system object */
DIR directory;					/* Directory object */
FILINFO file;				/* File information */

volatile uint8_t read1 = 1;

volatile uint16_t counter=0;

volatile uint8_t sync = 0;

volatile uint8_t write1 = 1;

volatile uint8_t rec_sample;

volatile uint32_t adc_val;

volatile uint32_t converted;
volatile uint32_t sent;

ISR(TIMER1_COMPA_vect){}
ISR(TIMER1_OVF_vect){}

ISR(TIMER0_COMPA_vect)
{
	if(sync == 1)
	{
		
		if(read1 == 1)
		{
			OCR1A = (unsigned char)Buff1[counter++];
			
			if(counter == 256)
			{
				read1 = 0;
				counter = 0;
			}
		}
		else if(read1 == 0)
		{
			OCR1A = (unsigned char)Buff2[counter++];
			if(counter == 256)
			{
				//PORTC = 0x40;
				read1 = 1;
				counter = 0;
			}
		}
	WUMETER();
	}

}

ISR(ADC_vect){
	// take sample value 
	adc_val = ADC;
	// normalize value
	adc_val -= 0x255;
	
	if (write1){
		Buff1[counter++] = adc_val;
		if (counter == WRITE_BUF_SIZE){
			counter = 0;
			converted ++;
			write1 = 0;
		}
	}
	else {
		Buff2[counter++] = adc_val;
		if (counter == WRITE_BUF_SIZE){
			counter = 0;
			converted ++;
			write1 = 1;
		}
	}
	
}

// =======================================================================================================================================

void Timer0_init(uint8_t div)
{
	TCCR0A = 0x02;
	switch(div){
		case 8:
			TCCR0B |= (1<<CS01);
			break;
		case 64:
			TCCR0B |= (1<<CS01) | (1<<CS00);
			break;
		default:
			TCCR0B |= (1<<CS00);
			break;
	}
	TIMSK0 = 0X02;
}

void Timer1_init(void)
{

	TCCR1A = _BV(COM1A1)|_BV(COM1A0)|_BV(WGM10);
	TCCR1B = _BV(WGM12)|_BV(CS10);
}

void ADC_init(void)
{ 
	ADMUX = (1<<REFS0);
	ADCSRA = (1<<ADEN) | (1<<ADSC) | (1<<ADATE) | (1<<ADIE) | (1<<ADPS2) | (1<<ADPS1);
	ADCSRB |= (1<<ADTS1) | (1<<ADTS0);
}

// =======================================================================================================================================


void controls_init(){
	// set control buttons
	CTRL_BUT_DDR &= ~(1<<CTRL_BUT_PLAY) & ~(1<<CTRL_BUT_REC);
	CTRL_BUT_PORT |=  (1<<CTRL_BUT_PLAY) | (1<<CTRL_BUT_REC);

	// set record led
	REC_LED_DDR |= (1<<REC_LED);
	
	TCCR2A=0x83;
	TCCR2B=0x01;

}



void filesystem_init(){
	_delay_ms(1000);
	// filesystem mount
	pf_mount(&file_system);
	// root directory opened
	_delay_ms(1000);
	pf_opendir(&directory, "");
	_delay_ms(1000);
}



/*---------------------------------------------------------*/

static
DWORD load_header (unsigned char *presc)	/* 0:Invalid format, 1:I/O error, >1:Number of samples */
{
	DWORD sz;


	if (pf_read(Buff1, 12, &rb)) return 1;	/* Load file header (12 bytes) */

	if (rb != 12 || LD_DWORD(Buff1+8) != FCC('W','A','V','E')) return 0;

	for (;;) {
		pf_read(Buff1, 8, &rb);			/* Get Chunk ID and size */
		if (rb != 8) return 0;
		sz = LD_DWORD(&Buff1[4]);		/* Chunk size */

		switch (LD_DWORD(&Buff1[0])) {	/* FCC */
		case FCC('f','m','t',' ') :					/* 'fmt ' chunk */
			if (sz > 100 || sz < 16) return 0;		/* Check chunk size */
			pf_read(Buff1, sz, &rb);					/* Get content */
			if (rb != sz) return 0;
			if (Buff1[0] != 1) return 0;				/* Check coding type (1) */
			if (Buff1[2] != 1 && Buff1[2] != 2) 		/* Check channels (1/2) */
				return 0;
			
			if (Buff1[14] != 8 && Buff1[14] != 16)	/* Check resolution (8/16) */
				return 0;
			
			if (F_CPU/8/LD_WORD(&Buff1[4]) < 256){
				*presc = 8;
				OCR0A = (BYTE)(F_CPU/8/LD_WORD(&Buff1[4]))-1;	/* Sampling freq */
			}
			else {
				*presc = 64;
				OCR0A = (BYTE)(F_CPU/64/LD_WORD(&Buff1[4]))-1;
			}
			
			break;

		case FCC('d','a','t','a') :				/* 'data' chunk (start to play) */
			return sz;

		case FCC('L','I','S','T') :				/* 'LIST' chunk (skip) */
		case FCC('f','a','c','t') :				/* 'fact' chunk (skip) */
			pf_lseek(file_system.fptr + sz);
			break;

		default :								/* Unknown chunk (error) */
			return 0;
		}
	}

	return 0;
}

void build_header(DWORD data_size){
	
	ST_DWORD(Buff1, 		FCC('R','I','F','F'));
	ST_DWORD(Buff1 + 4, 	504 + data_size);
	ST_DWORD(Buff1 + 8, 	FCC('W','A','V','E'));
	
	ST_DWORD(Buff1 + 12,	FCC('f','m','t',' '));
	ST_DWORD(Buff1 + 16,	16);
	ST_WORD(Buff1 + 20, 1);
	ST_WORD(Buff1 + 22, 1);
	ST_DWORD(Buff1 + 24, 16000);
	ST_DWORD(Buff1 + 28, 16000);
	ST_WORD(Buff1 + 32, 1);
	ST_WORD(Buff1 + 34, 8);
	
	ST_DWORD(Buff1 + 36, FCC('d','a','t','a'));
	ST_DWORD(Buff1 + 40, 468 + data_size);
	
	memset(Buff1 + 44, 0, 212);
	
}


/*-----------------------------------------------------------------------*/
/* Main                                                                  */



// ====================================================================================
// This is the part where I try to record and interact with the file system



void play_button_event(){
	DWORD sz;
	FRESULT res;
	unsigned char read2 = 8;

	uint16_t size = 256;
	
	///printf("playing");
	

	if (pf_open(file.fname) == FR_OK) {
	
		counter = 0;
		
		sz = load_header(&read2);		/* Load file header */
		if (sz < 256) return;
		
		sei();
		
		Timer0_init(read2);
		Timer1_init();
		
		// skip sector unalligned part
		rb = file_system.fptr % 512;
		pf_lseek(file_system.fptr + (512 - rb));
		sz -= rb;
	
		res = pf_read(Buff1, size, &rb);
		if (res != FR_OK) return; 
		
		sync = 1;
				
		do {
				
			read2 = read1;
			
			if(read1 == 1)	
				res = pf_read(Buff2, size, &rb);
			else
				res = pf_read(Buff1, size, &rb);
				
			if (res != FR_OK) { PORTC = 0x08; break; }
			
			
			while(read2 == read1);
			
		
		} while (rb == size && PLAY_BUT_PRESSED());	/* Repeat until all data read */
		
		sync = 0;
		TCCR1A &= ~(1<<COM1A1) & ~(1<<COM1A0);
		cli();
		
	}

	while(PLAY_BUT_PRESSED());
	_delay_ms(10);
	
	
	
}




void rec_button_event(){
	WORD written;
	unsigned char write2;
	DWORD data_size = 0;
	uint8_t tsent = 0;

	
	build_header(0);
	
	pf_open(file.fname);
	
	pf_write(Buff1, 256, &written);
	memset(Buff1, 0, 256);
	pf_write(Buff1, 256, &written);
	
	converted = 0;
	sent = 0;
	
	// set OCR0
	OCR0A = F_CPU/8/16000 - 1;
	counter = 0;
	write1 = 1;
	Timer0_init(8);
	ADC_init();
	SET_REC_LED();
	
	sei();
	
	write2 = write1;

	do {
		while (write2 == write1);
		
		if (write2)
			pf_write(Buff1, WRITE_BUF_SIZE, &written);
		else
			pf_write(Buff2, WRITE_BUF_SIZE, &written);
		
		sent ++;
		
		tsent++;
		if (tsent == 255){
			tsent = 0;
			data_size += 65280;
		}
		
		write2 = write1;
		
	} while(REC_BUT_PRESSED());
	
	cli();
	ADCSRA &= ~(1<<ADEN);
	pf_write(Buff1, 0, &written);
	
	pf_open(file.fname);
	
	build_header(data_size);
	pf_write(Buff1, 256, &written);
	memset(Buff1, 0, 256);
	pf_write(Buff1, 256, &written);
	pf_write(Buff1, 0, &written);
	
	
	while(REC_BUT_PRESSED());
	_delay_ms(10);
	CLR_REC_LED();
	
}
FILINFO next_file(FILINFO file){
	FILINFO result;
	do{
		pf_readdir(&directory, &result);
		if (result.fname[0] == 0){
			pf_opendir(&directory, "");
			pf_readdir(&directory, &result);
		}
	}	while (!strstr(result.fname, ".WAV"));
	return result;
			
}

int main(){
	
	// Initialization
	controls_init();

	filesystem_init();
	
	DDRD |= (1<<PD5);
	
	// display first file
	file = next_file(file);
	
	
	// open files
	while (1){
		if(PLAY_BUT_PRESSED()){
			play_button_event();
		}
		else if(REC_BUT_PRESSED()){
			rec_button_event();
		}
		OCR2A=0;
	}
	
	return 0;
	
}